# coding: utf8
from __future__ import division, print_function

import pytest
import numpy as np
import random
from numpy.testing import assert_allclose

from thinc.neural.optimizers import SGD
from thinc.neural._classes.affine import Affine


np.random.seed(2)
random.seed(0)


@pytest.fixture
def model():
    model = Affine(2, 2)
    return model


def test_init(model):
    assert model.nO == 2
    assert model.nI == 2
    assert model.W is not None
    assert model.b is not None


def test_predict_bias(model):
    input_ = model.ops.allocate((1, model.nI))
    target_scores = model.ops.allocate((1, model.nO))
    scores = model(input_)
    assert_allclose(scores[0], target_scores[0])

    # Set bias for class 0
    model.b[0] = 2.0
    target_scores[0, 0] = 2.0
    scores = model(input_)
    assert_allclose(scores, target_scores)

    # Set bias for class 1
    model.b[1] = 5.0
    target_scores[0, 1] = 5.0
    scores = model(input_)
    assert_allclose(scores, target_scores)


@pytest.mark.parametrize(
    "X,expected",
    [
        (np.asarray([0.0, 0.0], dtype="f"), [0.0, 0.0]),
        (np.asarray([1.0, 0.0], dtype="f"), [1.0, 0.0]),
        (np.asarray([0.0, 1.0], dtype="f"), [0.0, 1.0]),
        (np.asarray([1.0, 1.0], dtype="f"), [1.0, 1.0]),
    ],
)
def test_predict_weights(X, expected):
    W = np.asarray([1.0, 0.0, 0.0, 1.0], dtype="f").reshape((2, 2))
    bias = np.asarray([0.0, 0.0], dtype="f")
    print(W)
    print(bias)

    model = Affine(W.shape[0], W.shape[1])
    model.W[:] = W
    model.b[:] = bias

    scores = model.predict_one(X)
    print("scores", scores, "expected", expected)
    assert_allclose(scores, expected)


def test_update():
    W = np.asarray([1.0, 0.0, 0.0, 1.0], dtype="f").reshape((2, 2))
    bias = np.asarray([0.0, 0.0], dtype="f")

    model = Affine(2, 2)
    model.W[:] = W
    model.b[:] = bias
    sgd = SGD(model.ops, 1.0, L2=0.0)
    sgd.averages = None

    ff = np.asarray([[0.0, 0.0]], dtype="f")
    tf = np.asarray([[1.0, 0.0]], dtype="f")
    ft = np.asarray([[0.0, 1.0]], dtype="f")  # noqa: F841
    tt = np.asarray([[1.0, 1.0]], dtype="f")  # noqa: F841

    # ff, i.e. 0, 0
    scores, finish_update = model.begin_update(ff)
    assert_allclose(scores[0, 0], scores[0, 1])
    # Tell it the answer was 'f'
    gradient = np.asarray([[-1.0, 0.0]], dtype="f")
    finish_update(gradient, sgd)

    assert model.b[0] == 1.0
    assert model.b[1] == 0.0
    # Unchanged -- input was zeros, so can't get gradient for weights.
    assert model.W[0, 0] == 1.0
    assert model.W[0, 1] == 0.0
    assert model.W[1, 0] == 0.0
    assert model.W[1, 1] == 1.0

    # tf, i.e. 1, 0
    scores, finish_update = model.begin_update(tf)
    # Tell it the answer was 'T'
    gradient = np.asarray([[0.0, -1.0]], dtype="f")
    finish_update(gradient, sgd)

    assert model.b[0] == 1.0
    assert model.b[1] == 1.0
    # Gradient for weights should have been outer(gradient, input)
    # so outer([0, -1.], [1., 0.])
    # =  [[0., 0.], [-1., 0.]]
    assert model.W[0, 0] == 1.0 - 0.0
    assert model.W[0, 1] == 0.0 - 0.0
    assert model.W[1, 0] == 0.0 - -1.0
    assert model.W[1, 1] == 1.0 - 0.0



# coding: utf8
from __future__ import unicode_literals

from thinc.neural._classes.maxout import Maxout
from thinc.api import chain


def test_simple_model_roundtrip_bytes():
    model = Maxout(5, 10, pieces=2)
    model.b += 1
    data = model.to_bytes()
    model.b -= 1
    model = model.from_bytes(data)
    assert model.b[0, 0] == 1


def test_multi_model_roundtrip_bytes():
    model = chain(Maxout(5, 10, pieces=2), Maxout(2, 3))
    model._layers[0].b += 1
    model._layers[1].b += 2
    data = model.to_bytes()
    model._layers[0].b -= 1
    model._layers[1].b -= 2
    model = model.from_bytes(data)
    assert model._layers[0].b[0, 0] == 1
    assert model._layers[1].b[0, 0] == 2


def test_multi_model_load_missing_dims():
    model = chain(Maxout(5, 10, pieces=2), Maxout(2, 3))
    model._layers[0].b += 1
    model._layers[1].b += 2
    data = model.to_bytes()

    model2 = chain(Maxout(5), Maxout())
    model2 = model2.from_bytes(data)
    assert model2._layers[0].b[0, 0] == 1
    assert model2._layers[1].b[0, 0] == 2



# coding: utf8
from __future__ import unicode_literals, print_function

import pytest

from thinc.neural.vec2vec import Model, ReLu, Softmax
from thinc.neural._classes.feed_forward import FeedForward
from thinc.neural._classes.batchnorm import BatchNorm
from thinc.neural._classes.elu import ELU
from thinc.neural._classes.maxout import Maxout
from thinc.neural.ops import NumpyOps
from thinc.api import clone, chain
from thinc.neural.util import to_categorical
from thinc.extra import datasets


@pytest.fixture(scope="module")
def mnist():
    train_data, dev_data, _ = datasets.mnist()
    train_X, train_y = NumpyOps().unzip(train_data)
    dev_X, dev_y = NumpyOps().unzip(dev_data)
    dev_y = to_categorical(dev_y, nb_classes=10)
    train_y = to_categorical(dev_y, nb_classes=10)
    return (train_X[:1000], train_y[:1000]), (dev_X, dev_y)


@pytest.fixture(scope="module")
def train_X(mnist):
    (train_X, train_y), (dev_X, dev_y) = mnist
    return train_X


@pytest.fixture(scope="module")
def train_y(mnist):
    (train_X, train_y), (dev_X, dev_y) = mnist
    return train_y


@pytest.fixture(scope="module")
def dev_X(mnist):
    (train_X, train_y), (dev_X, dev_y) = mnist
    return dev_X


@pytest.fixture(scope="module")
def dev_y(mnist):
    (train_X, train_y), (dev_X, dev_y) = mnist
    return dev_y


def create_relu_softmax(depth, width):
    with Model.define_operators({"*": clone, ">>": chain}):
        model = ReLu(width, 784) >> Softmax(10, width)
    return model


def create_relu_batchnorm_softmax(depth, width):
    with Model.define_operators({"*": clone, ">>": chain}):
        model = BatchNorm(ReLu(width, 784)) >> Softmax(10, width)
    return model


def create_elu_maxout_softmax(depth, width):
    with Model.define_operators({">>": chain}):
        model = ELU(width) >> Maxout(5) >> Softmax()
    return model


@pytest.fixture(
    params=[
        create_relu_softmax,
        create_elu_maxout_softmax,
        create_relu_batchnorm_softmax,
    ]
)
def create_model(request):
    return request.param


@pytest.mark.slow
@pytest.mark.parametrize(("depth", "width", "nb_epoch"), [(2, 8, 5)])
def test_small_end_to_end(
    depth, width, nb_epoch, create_model, train_X, train_y, dev_X, dev_y
):
    model = create_model(depth, width)
    assert isinstance(model, FeedForward)
    losses = []
    with model.begin_training(train_X, train_y) as (trainer, optimizer):
        trainer.each_epoch.append(lambda: print(model.evaluate(dev_X, dev_y)))
        trainer.nb_epoch = nb_epoch
        for X, y in trainer.iterate(train_X, train_y):
            X = model.ops.asarray(X)
            y = model.ops.asarray(y)
            yh, backprop = model.begin_update(X, drop=trainer.dropout)
            backprop(yh - y, optimizer)
            loss = ((yh - y) ** 2).sum()
            losses.append(loss)
    assert losses[-1] < losses[0]



# coding: utf8
from __future__ import unicode_literals

import numpy


try:
    from cupy import get_array_module
except ImportError:

    def get_array_module(*a, **k):
        return numpy


def categorical_crossentropy(scores, labels):
    xp = get_array_module(scores)
    target = xp.zeros(scores.shape, dtype="float32")
    loss = 0.0
    for i in range(len(labels)):
        target[i, int(labels[i])] = 1.0
        loss += (1.0 - scores[i, int(labels[i])]) ** 2
    return scores - target, loss


def L1_distance(vec1, vec2, labels, margin=0.2):
    xp = get_array_module(vec1)
    dist = xp.abs(vec1 - vec2).sum(axis=1)
    loss = (dist > margin) - labels
    return (vec1 - vec2) * loss, (vec2 - vec1) * loss, loss



# coding: utf8
from __future__ import unicode_literals

try:
    # Python >= 3.3
    from collections.abc import Sequence, Sized, Iterable, Callable
except ImportError:
    # Python < 3.3
    from collections import Sequence, Sized, Iterable, Callable
from numpy import ndarray

from .compat import integer_types
from .extra import wrapt
from .exceptions import UndefinedOperatorError, DifferentLengthError
from .exceptions import ExpectedTypeError, ShapeMismatchError
from .exceptions import OutsideRangeError


def is_docs(arg_id, args, kwargs):
    from spacy.tokens.doc import Doc

    docs = args[arg_id]
    if not isinstance(docs, Sequence):
        raise ExpectedTypeError(type(docs), ["Sequence"])
    if not isinstance(docs[0], Doc):
        raise ExpectedTypeError(type(docs[0]), ["spacy.tokens.doc.Doc"])


def equal_length(*args):
    """Check that arguments have the same length.
    """
    for i, arg in enumerate(args):
        if not isinstance(arg, Sized):
            raise ExpectedTypeError(arg, ["Sized"])
        if i >= 1 and len(arg) != len(args[0]):
            raise DifferentLengthError(args, arg)


def equal_axis(*args, **axis):
    """Check that elements have the same dimension on specified axis.
    """
    axis = axis.get("axis", -1)
    for i, arg in enumerate(args):
        if not isinstance(arg, ndarray):
            raise ExpectedTypeError(arg, ["ndarray"])
        if axis >= 0 and (axis + 1) < arg.shape[axis]:
            raise ShapeMismatchError(arg.shape[axis], axis, [])
        if i >= 1 and arg.shape[axis] != args[0].shape[axis]:
            lengths = [a.shape[axis] for a in args]
            raise DifferentLengthError(lengths, arg)


def has_shape(shape):
    """Check that a particular argument is an array with a given shape. The
    shape may contain string attributes, which will be fetched from arg0 to
    the function (usually self).
    """

    def has_shape_inner(arg_id, args, kwargs):
        self = args[0]
        arg = args[arg_id]
        if not hasattr(arg, "shape"):
            raise ExpectedTypeError(arg, ["array"])
        shape_values = []
        for dim in shape:
            if not isinstance(dim, integer_types):
                dim = getattr(self, dim, None)
            shape_values.append(dim)
        if len(shape) != len(arg.shape):
            raise ShapeMismatchError(arg.shape, tuple(shape_values), shape)
        for i, dim in enumerate(shape_values):
            # Allow underspecified dimensions
            if dim is not None and arg.shape[i] != dim:
                raise ShapeMismatchError(arg.shape, shape_values, shape)

    return has_shape_inner


def is_shape(arg_id, args, func_kwargs, **kwargs):
    arg = args[arg_id]
    if not isinstance(arg, Iterable):
        raise ExpectedTypeError(arg, ["iterable"])
    for value in arg:
        if not isinstance(value, integer_types) or value < 0:
            raise ExpectedTypeError(arg, ["valid shape (positive ints)"])


def is_sequence(arg_id, args, kwargs):
    arg = args[arg_id]
    if not isinstance(arg, Iterable) and not hasattr(arg, "__getitem__"):
        raise ExpectedTypeError(arg, ["iterable"])


def is_float(arg_id, args, func_kwargs, **kwargs):
    arg = args[arg_id]
    if not isinstance(arg, float):
        raise ExpectedTypeError(arg, ["float"])
    if "min" in kwargs and arg < kwargs["min"]:
        raise OutsideRangeError(arg, kwargs["min"], ">=")
    if "max" in kwargs and arg > kwargs["max"]:
        raise OutsideRangeError(arg, kwargs["max"], "<=")


def is_int(arg_id, args, func_kwargs, **kwargs):
    arg = args[arg_id]
    if not isinstance(arg, integer_types):
        raise ExpectedTypeError(arg, ["int"])
    if "min" in kwargs and arg < kwargs["min"]:
        raise OutsideRangeError(arg, kwargs["min"], ">=")
    if "max" in kwargs and arg > kwargs["max"]:
        raise OutsideRangeError(arg, kwargs["max"], "<=")


def is_array(arg_id, args, func_kwargs, **kwargs):
    arg = args[arg_id]
    if not isinstance(arg, ndarray):
        raise ExpectedTypeError(arg, ["ndarray"])


def is_int_array(arg_id, args, func_kwargs, **kwargs):
    arg = args[arg_id]
    if not isinstance(arg, ndarray) or "i" not in arg.dtype.kind:
        raise ExpectedTypeError(arg, ["ndarray[int]"])


def operator_is_defined(op):
    @wrapt.decorator
    def checker(wrapped, instance, args, kwargs):
        if instance is None:
            instance = args[0]
        if instance is None:
            raise ExpectedTypeError(instance, ["Model"])
        if op not in instance._thread_local.operators:
            raise UndefinedOperatorError(op, instance, args[0], instance._thread_local.operators)
        else:
            return wrapped(*args, **kwargs)

    return checker


def arg(arg_id, *constraints):
    @wrapt.decorator
    def checked_function(wrapped, instance, args, kwargs):
        # for partial functions or other C-compiled functions
        if not hasattr(wrapped, "checks"):  # pragma: no cover
            return wrapped(*args, **kwargs)
        if instance is not None:
            fix_args = [instance] + list(args)
        else:
            fix_args = list(args)
        for arg_id, checks in wrapped.checks.items():
            for check in checks:
                if not isinstance(check, Callable):
                    raise ExpectedTypeError(check, ["Callable"])
                check(arg_id, fix_args, kwargs)
        return wrapped(*args, **kwargs)

    def arg_check_adder(func):
        if hasattr(func, "checks"):
            func.checks.setdefault(arg_id, []).extend(constraints)
            return func
        else:
            wrapped = checked_function(func)
            wrapped.checks = {arg_id: list(constraints)}
            return wrapped

    return arg_check_adder


def args(*constraints):
    @wrapt.decorator
    def arg_check_adder(wrapped, instance, args, kwargs):
        for check in constraints:
            if not isinstance(check, Callable):
                raise ExpectedTypeError(check, ["Callable"])
            check(*args)
        return wrapped(*args, **kwargs)

    return arg_check_adder



import catalogue


class registry(object):
    optimizers = catalogue.create("thinc", "optimizers", entry_points=True)
    schedules = catalogue.create("thinc", "schedules", entry_points=True)
    layers = catalogue.create("thinc", "layers", entry_points=True)

    @classmethod
    def get(cls, name, key):
        if not hasattr(cls, name):
            raise ValueError("Unknown registry: %s" % name)
        reg = getattr(cls, name)
        func = reg.get(key)
        if func is None:
            raise ValueError("Could not find %s in %s" % (name, key))
        return func

    @classmethod
    def make_optimizer(name, args, kwargs):
        func = cls.optimizers.get(name)
        return func(*args, **kwargs)

    @classmethod
    def make_schedule(name, args, kwargs):
        func = cls.schedules.get(name)
        return func(*args, **kwargs)

    @classmethod
    def make_initializer(name, args, kwargs):
        func = cls.initializers.get(name)
        return func(*args, **kwargs)

    @classmethod
    def make_layer(cls, name, args, kwargs):
        func = cls.layers.get(name)
        return func(*args, **kwargs)

    @classmethod
    def make_combinator(cls, name, args, kwargs):
        func = cls.combinators.get(name)
        return func(*args, **kwargs)

    @classmethod
    def make_transform(cls, name, args, kwargs):
        func = cls.transforms.get(name)
        return func(*args, **kwargs)

    @classmethod
    def make_from_config(cls, config, id_start="@"):
        """Unpack a config dictionary, creating objects from the registry 
        recursively.
        """
        id_keys = [key for key in config.keys() if key.startswith(id_start)]
        if len(id_keys) >= 2:
            raise ValueError("Multiple registry keys in config: %s" % id_keys)
        elif len(id_keys) == 0:
            # Recurse over subdictionaries, filling in values.
            filled = {}
            for key, value in config.items():
                if isinstance(value, dict):
                    filled[key] = cls.make_from_config(value, id_start=id_start)
                else:
                    filled[key] = value
            return filled
        else:
            getter = cls.get(id_keys[0].replace(id_start, ""), config[id_keys[0]])
            args = []
            kwargs = {}
            for key, value in config.items():
                if isinstance(value, dict):
                    value = cls.make_from_config(value, id_start=id_start)
                if isinstance(key, int) or key.isdigit():
                    args.append((int(key), value))
                elif not key.startswith(id_start):
                    kwargs[key] = value
            args = [value for key, value in sorted(args)]
            return getter(*args, **kwargs)



# coding: utf8
from __future__ import unicode_literals

import numpy

from thinc.neural._classes.hash_embed import HashEmbed


def test_init():
    model = HashEmbed(64, 1000)
    assert model.nV == 1000
    assert model.nO == 64
    assert model.vectors.shape == (1000, 64)
    # assert model.word_weights.shape == (1000,)


def test_seed_changes_bucket():
    model1 = HashEmbed(64, 1000, seed=2)
    model2 = HashEmbed(64, 1000, seed=1)
    arr = numpy.ones((1,), dtype="uint64")
    vector1 = model1(arr)
    vector2 = model2(arr)
    assert vector1.sum() != vector2.sum()



""" A visualizer module for Thinc """

try:
    import seaborn
    import matplotlib.pyplot as plt
except ImportError:
    pass


def visualize_attention(x, y, weights, layer="Encoder", self_attn=True):
    """
        Visualize self/outer attention
        Args:
            x: sentence
            y: sentence
            weights: (nH, nL, nL)
    """

    def heatmap(x, y, data, ax):
        seaborn.heatmap(
            data,
            square=True,
            xticklabels=y,
            yticklabels=x,
            vmin=0.0,
            vmax=1.0,
            cbar_kws=dict(use_gridspec=False, location="top"),
            ax=ax,
        )

    num = min(weights.shape[0], 4)
    fig, axs = plt.subplots(1, num)
    attn_type = "self attention" if self_attn else "outer attention"
    fig.suptitle("{} {} for all the heads".format(layer, attn_type))
    if len(weights.shape) == 3:
        for i in range(num):
            heatmap(x, y, weights[i], axs[i])
    else:
        raise ValueError("Wrong input weights dimensions")
    plt.show()



# coding: utf8
from __future__ import unicode_literals

from .model import Model
from ... import describe


def _init_to_one(W, ops):
    W.fill(1.0)


def _run_child_hooks(model, X, y=None):
    for hook in model.child.on_data_hooks:
        hook(model.child, X, y)


@describe.on_data(_run_child_hooks)
@describe.attributes(
    G=describe.Weights("Scaling vector", lambda obj: (obj.nO,), _init_to_one),
    b=describe.Biases("Bias vector", lambda obj: (obj.nO,)),
    d_G=describe.Gradient("G"),
    d_b=describe.Gradient("b"),
    m=describe.Weights("Means", lambda obj: (obj.nO,)),
    v=describe.Weights("Variance", lambda obj: (obj.nO,), _init_to_one),
)
class BatchNorm(Model):
    name = "batchnorm"

    def __init__(self, child, **kwargs):
        self.child = child
        self._layers = [child]
        if "nO" in kwargs:
            self.nO = kwargs["nO"]
        elif getattr(child, "nO", None):
            self.nO = child.nO
        self.nr_upd = 0
        self.eps = kwargs.get("eps", 1e-5)
        self.alpha = self.ops.xp.asarray([0.1], dtype="float32")
        self.rmax = kwargs.get("rmax", 3.0)
        self.dmax = kwargs.get("dmax", 5.0)
        Model.__init__(self, **kwargs)

    def predict(self, X):
        X = self.child.predict(X)
        Xh = _forward(self.ops, X, self.m, self.v + self.eps)
        y = Xh * self.G + self.b
        return y

    def begin_update(self, X, drop=0.0):
        if drop is None:
            return self.predict(X), None
        assert X.dtype == "float32"
        X, backprop_child = self.child.begin_update(X, drop=0.0)
        N, mu, var = _get_moments(self.ops, X)
        var += self.eps
        r = self.ops.xp.clip(var / self.v, 1.0 / self.rmax, self.rmax)
        d = self.ops.xp.clip((mu - self.m) / self.v, -self.dmax, self.dmax)
        self.nr_upd += 1

        # I'm not sure this is the best thing to do --
        # Should we consider a sample be the instance, or the batch?
        # If we want the variance of the inputs it should be like:
        """
        diff = X - self.m
        incr = (1-alpha) * diff
        self.m += incr.mean(axis=0)
        self.v += (diff * incr).mean(axis=0)
        self.v *= alpha
        """
        self.m += self.alpha * (mu - self.m)
        self.v += self.alpha * (var - self.v)
        Xhat = _forward(self.ops, X, mu, var)
        Xhat *= r
        Xhat += d

        y, backprop_rescale = self._begin_update_scale_shift(Xhat)

        def finish_update(dy, sgd=None):
            dy = backprop_rescale(dy, sgd)
            dist, sum_dy, sum_dy_dist = _get_d_moments(self.ops, dy, X, mu)
            d_xhat = N * dy - sum_dy - dist * (1.0 / var) * sum_dy_dist
            d_xhat *= var ** (-1.0 / 2)
            d_xhat /= N
            return backprop_child(d_xhat, sgd)

        if drop is not None:
            drop *= getattr(self.child, "drop_factor", 1.0)
        y, bp_dropout = self.ops.dropout(y, drop)
        assert y.dtype == "float32"
        return y, bp_dropout(finish_update)

    def _begin_update_scale_shift(self, input__BI):
        def finish_update(gradient__BI, sgd=None):
            self.d_b += gradient__BI.sum(axis=0)
            d_G = self.d_G
            d_G += (gradient__BI * input__BI).sum(axis=0)
            if sgd is not None:
                sgd(self._mem.weights, self._mem.gradient, key=self.id)
            return gradient__BI * self.G

        return input__BI * self.G + self.b, finish_update


def _get_moments(ops, X):
    mu = X.mean(axis=0)
    var = X.var(axis=0) + 1e-08
    return ops.asarray([X.shape[0]], dtype="float32"), mu, var


def _get_d_moments(ops, dy, X, mu):
    dist = X - mu
    return dist, ops.xp.sum(dy, axis=0), ops.xp.sum(dy * dist, axis=0)


def _forward(ops, X, mu, var):
    return (X - mu) * var ** (-1.0 / 2.0)



# coding: utf8
from __future__ import unicode_literals
import sys


try:
    basestring = basestring
except NameError:
    basestring = str


try:
    from StringIO import StringIO as BytesIO
except ImportError:
    from io import BytesIO  # noqa: F401


# See: https://github.com/benjaminp/six/blob/master/six.py
is_python2 = sys.version_info[0] == 2
is_python3 = sys.version_info[0] == 3
is_python_pre_3_5 = is_python2 or (is_python3 and sys.version_info[1] < 5)

if is_python3:
    integer_types = (int,)
    string_types = (str,)
    from urllib.error import URLError, HTTPError
    from urllib.request import urlopen, urlretrieve
else:
    integer_types = (int, long)
    string_types = (basestring,)
    from urllib2 import URLError, HTTPError
    from urllib2 import urlopen
    from urllib import urlretrieve



# coding: utf8
from __future__ import unicode_literals

import contextlib
from ..compat import BytesIO
from ..neural._classes.model import Model

try:
    import cupy
except ImportError:
    cupy = None

try:
    import torch.autograd
    import torch.optim
    import torch
    import torch.utils.dlpack
    from torch.nn import Module as PyTorchModule
except ImportError:
    torch = None
    PyTorchModule = None


def xp2torch(xp_tensor):
    if hasattr(xp_tensor, "toDlpack"):
        return torch.utils.dlpack.from_dlpack(xp_tensor.toDlpack())
    else:
        return torch.from_numpy(xp_tensor)


def torch2xp(torch_tensor):
    if torch_tensor.is_cuda:
        return cupy.fromDlpack(torch.utils.dlpack.to_dlpack(torch_tensor))
    else:
        return torch_tensor.detach().numpy()


class PyTorchWrapper(Model):
    """Wrap a PyTorch model, so that it has the same API as Thinc models.
    To optimize the model, you'll need to create a PyTorch optimizer and call
    optimizer.step() after each batch --- see examples/wrap_pytorch.py
    """

    def __init__(self, model):
        Model.__init__(self)
        self._model = model
        self._optimizer = None

    def prepare_input(self, x_data, is_update=True):
        if isinstance(x_data, (list, tuple)):
            x_var = [torch.autograd.Variable(xp2torch(x), requires_grad=is_update)
                     for x in x_data]
            return tuple(x_var), {}
        else:
            x_var = torch.autograd.Variable(xp2torch(x_data), requires_grad=is_update)
            return (x_var,), {}

    def prepare_output(self, y_var):
        if isinstance(y_var, (list, tuple)):
            return tuple([torch2xp(y) for y in y_var])
        else:
            return torch2xp(y_var)

    def prepare_backward_input(self, dy_data, y_var):
        if isinstance(dy_data, (list, tuple)):
            dy_var = [xp2torch(dy) for dy in dy_data]
            return y_var, {"grad_tensors": dy_var}
        else:
            dy_var = xp2torch(dy_data)
            return (y_var,), {"grad_tensors": (dy_var,)}

    def prepare_backward_output(self, x_args, x_kwargs):
        x_var = x_args[0]
        return torch2xp(x_var.grad)

    def predict(self, x_data):
        self._model.eval()
        x_args, x_kwargs = self.prepare_input(x_data, is_update=False)
        with torch.no_grad():
            y_var = self._model(*x_args, **x_kwargs)
        self._model.train()
        return self.prepare_output(y_var)

    def begin_update(self, x_data, drop=0.0):
        """Return the output of the wrapped PyTorch model for the given input,
        along with a callback to handle the backward pass.
        """
        if drop is None:
            return self.predict(x_data), None
        self._model.train()
        fwd_args, fwd_kwargs = self.prepare_input(x_data, is_update=True)
        y_var = self._model(*fwd_args, **fwd_kwargs)
        y = self.prepare_output(y_var)

        def backward_pytorch(dy_data, sgd=None):
            d_args, d_kwargs = self.prepare_backward_input(dy_data, y_var)
            torch.autograd.backward(*d_args, **d_kwargs)
            if sgd is not None:
                if self._optimizer is None:
                    self._optimizer = self._create_optimizer(sgd)
                if getattr(sgd, "max_grad_norm", None):
                    torch.nn.utils.clip_grad_norm_(
                        self._model.parameters(),
                        sgd.max_grad_norm)
                self._optimizer.step()
                self._optimizer.zero_grad()
            return self.prepare_backward_output(fwd_args, fwd_kwargs)

        return y, backward_pytorch

    def _create_optimizer(self, sgd):
        params = self._model.parameters()
        if sgd.b1 != 0 and sgd.b2 != 0:
            optimizer = torch.optim.Adam(params, lr=sgd.alpha, betas=(sgd.b1, sgd.b2))
        elif sgd.b2 == 0:
            optimizer = torch.optim.SGD(params, lr=sgd.alpha, momentum=sgd.b1)
        else:
            raise NotImplementedError
        return optimizer

    @contextlib.contextmanager
    def use_params(self, params):
        key_prefix = "pytorch_{id_}_".format(id_=self.id)
        state_dict = {}
        for k, v in params.items():
            if hasattr(k, "startswith") and k.startswith(key_prefix):
                state_dict[k.replace(key_prefix, "")] = xp2torch(v)
        if state_dict:
            backup = {k: v.clone() for k, v in self._model.state_dict().items()}
            self._model.load_state_dict(state_dict)
            yield
            self._model.load_state_dict(backup)
        else:
            yield

    def _update_pytorch_averages(self, sgd, *, init_steps=1):
        if getattr(sgd, "averages", None) is None:
            return
        # Collect parameters if we don't have them
        for name, param in self._model.state_dict().items():
            key = "pytorch_{id_}_{name}".format(id_=self.id, name=name)
            sgd.nr_update[key] += 1
            xp_param = torch2xp(param)
            if key in sgd.averages:
                self.ops.update_averages(
                    sgd.averages[key], xp_param, sgd.nr_update[key]
                )
            else:
                sgd.averages[key] = xp_param.copy()
                sgd.nr_update[key] = init_steps

    def to_disk(self, path):
        torch.save(self._model.state_dict(), str(path))

    def from_disk(self, path):
        if self.ops.device == "cpu":
            map_location = "cpu"
        else:
            device_id = torch.cuda.current_device()
            map_location = "cuda:%d" % device_id
        self._model.load_state_dict(torch.load(path, map_location=map_location))
        self._model.to(map_location)

    def to_bytes(self):
        filelike = BytesIO()
        torch.save(self._model.state_dict(), filelike)
        filelike.seek(0)
        return filelike.getvalue()

    def from_bytes(self, data):
        filelike = BytesIO(data)
        filelike.seek(0)
        if self.ops.device == "cpu":
            map_location = "cpu"
        else:
            device_id = torch.cuda.current_device()
            map_location = "cuda:%d" % device_id
        self._model.load_state_dict(torch.load(filelike, map_location=map_location))
        self._model.to(map_location)

    def to_gpu(self, device_num):
        self._model.cuda(device_num)

    def to_cpu(self):
        self._model.cpu()

    def resize_output(self, new_dim):
        # self.weight = nn.Parameter(F.pad(self.weight, ...)) # add classes
        # self.weight = nn.Parameter(F.pad(model.weight, ...)) # add classes
        raise NotImplementedError

    def resize_input(self):
        raise NotImplementedError


class PyTorchWrapperRNN(PyTorchWrapper):
    """Wrap a PyTorch RNN model"""

    def prepare_input(self, inputs, is_update=False):
        if isinstance(inputs, tuple):
            x_data, h_0 = inputs
        else:
            x_data = inputs
            h_0 = None
        x_var = torch.autograd.Variable(xp2torch(x_data), requires_grad=is_update)
        return (x_var, h_0), {}

    def prepare_output(self, torch_outputs):
        y_var, h_n = torch_outputs
        return torch2xp(y_var), h_n

    def prepare_backward_input(self, dy_data, y_var):
        dy, _ = dy_data
        dy_var = xp2torch(dy)
        y_var, _ = y_var
        return (y_var,), {"grad_tensors": (dy_var,)}

    def prepare_backward_output(self, x_args, x_kwargs):
        x_var, _ = x_args
        return torch2xp(x_var.grad)
